######## Imports ########
#### Standard Library ####
import time
#### Third Party ####
import numpy as np
from PIL import Image
import matplotlib
from matplotlib import pyplot as plt
#### CSRK ####
from csrk import HybridGP

######## Setup ########
# Image location
FNAME_ORIGINAL = "originals/Nighthawks_2993_5180.jpg"
FNAME_OUT = "csrk_Nighthawks.jpg"
# GP stuff
train_width = 800
#train_width = 480
sample_width = 3996
#sample_width = 1920

## Hyperparameters ##
# For reconstructions of hand-drawn images that have been scanned,
#   resolution trumps all. Keep scale at 2.0 and order at 1.
scale = np.full(2,2.0)
whitenoise = 0.
order = 1

######## Load Image ########
# Load image pixel data
tic = time.perf_counter()
raw = Image.open(FNAME_ORIGINAL)
raw_YCbCr = raw.convert('YCbCr')
pixels_YCbCr = raw_YCbCr.load()
width, height = raw_YCbCr.size
toc = time.perf_counter()
print(f"Pillow initialization time: {toc-tic:.6f} s")

######## Generate Training Data for GP ########
tic = time.perf_counter()
# Define what we want to do wth the GP
train_height = int(train_width * (height / width))
train_height = min(train_height, height)
train_width = min(train_width, width)
sample_height = int(sample_width * (height / width))

# Define training x data
training_x_loc = np.arange(train_width)
training_y_loc = np.arange(train_height)
training_x_loc, training_y_loc = np.meshgrid(training_x_loc, training_y_loc)
training_x_loc = training_x_loc.flatten()
training_y_loc = training_y_loc.flatten()
X = np.column_stack([training_x_loc, training_y_loc])
# Define Y_err
Y_err = np.zeros(X.shape[0])

# Define sample data
eval_x_loc = np.arange(sample_width) * (train_width / sample_width)
eval_y_loc = np.arange(sample_height) * (train_height / sample_height)
eval_x_loc, eval_y_loc = np.meshgrid(eval_x_loc, eval_y_loc)
eval_x_loc = eval_x_loc.flatten()
eval_y_loc = eval_y_loc.flatten()
X_eval = np.column_stack((eval_x_loc, eval_y_loc)).astype(float)
    
# Define block width / height
block_width = int(np.floor(width / train_width))
block_height = int(np.floor(height / train_height))
block_size = block_height * block_height

# Define block "stamp"
block_x = np.arange(block_width).astype(int)
block_y = np.arange(block_height).astype(int)
block_x, block_y = np.meshgrid(block_x, block_y)
block_x = block_x.flatten()
block_y = block_y.flatten()

# Define Y data
Y_uv        = []
Y_Cb        = []
Y_Cr        = []
for (i, j) in X:
    x_start = i * block_width
    y_start = j * block_height
    x = block_x + x_start
    y = block_y + y_start
    lums, Cbs, Crs = 0, 0, 0
    for (xi, yi) in zip(x,y):
        lum, cb, cr = pixels_YCbCr[xi, yi]
        lums += lum
        Cbs += cb
        Crs += cr
    Y_uv.append(    lums    / block_size)
    Y_Cb.append(    Cbs     / block_size)
    Y_Cr.append(    Crs     / block_size)

Y_uv = np.asarray(Y_uv)
Y_Cb = np.asarray(Y_Cb)
Y_Cr = np.asarray(Y_Cr)
# Change X to float
X = X.astype(float)

toc = time.perf_counter()
print(f"Setting up training data time: {toc-tic:.6f} s!")

#### Train Gaussian Processes ####

# uv
print(f"Training lum GP on {train_width} by {train_height} grid!")
tic = time.perf_counter()
gp_uv = HybridGP(
    X,
    Y_uv,
    Y_err,
    scale,
    whitenoise,
    order,
)
toc = time.perf_counter()
print(f"time: {toc-tic:.6f} s")
# Cb
print(f"Training Cb GP on {train_width} by {train_height} grid!")
tic = time.perf_counter()
gp_Cb = HybridGP(
    X,
    Y_Cb,
    Y_err,
    scale,
    whitenoise,
    order,
)
toc = time.perf_counter()
print(f"time: {toc-tic:.6f} s")
# Cr
print(f"Training Cr GP on {train_width} by {train_height} grid!")
tic = time.perf_counter()
gp_Cr = HybridGP(
    X,
    Y_Cr,
    Y_err,
    scale,
    whitenoise,
    order,
)
toc = time.perf_counter()
print(f"time: {toc-tic:.6f} s")

#### Evaluate sample data ####

# Eval uv
print(f"Evaluating uv GP on {sample_width} by {sample_height} grid!")
tic = time.perf_counter()
y_eval_uv = gp_uv.predict_mean(X_eval)
toc = time.perf_counter()
print(f"time: {toc-tic:.6f} s")

# Eval Cr
print(f"Evaluating Cr GP on {sample_width} by {sample_height} grid!")
tic = time.perf_counter()
y_eval_Cb = gp_Cb.predict_mean(X_eval)
toc = time.perf_counter()
print(f"time: {toc-tic:.6f} s")

# Eval Cb
print(f"Evaluating Cb GP on {sample_width} by {sample_height} grid!")
tic = time.perf_counter()
y_eval_Cr = gp_Cr.predict_mean(X_eval)
toc = time.perf_counter()
print(f"time: {toc-tic:.6f} s")

y_eval_uv = y_eval_uv.reshape((sample_height,sample_width)).astype('u1')
y_eval_Cb = y_eval_Cb.reshape((sample_height,sample_width)).astype('u1')
y_eval_Cr = y_eval_Cr.reshape((sample_height,sample_width)).astype('u1')
y_eval = np.stack([y_eval_uv, y_eval_Cb, y_eval_Cr], axis=2)

#### Manage output image ####
# Create new image in Pillow
new_image = Image.fromarray(
    y_eval,
    "YCbCr"
)
# Save image
new_image.save(FNAME_OUT)



def test_Nighthawks():
    import time
    import numpy as np
    from PIL import Image
    import matplotlib
    from matplotlib import pyplot as plt
    from csrk import HybridGP

    # Load image pixel data
    tic = time.perf_counter()
    raw = Image.open("originals/Nighthawks_1600_870.jpg")
    pixels = raw.load()
    width, height = raw.size
    toc = time.perf_counter()
    print(f"Pillow initialization time: {toc-tic:.6f} s")

    tic = time.perf_counter()
    # Define what we want to do wth the GP
    train_width = 480
    train_height = int(train_width * (height / width))
    train_height = min(train_height, height)
    train_width = min(train_width, width)
    sample_width = 1920
    sample_height = int(sample_width * (height / width))
    scale = np.full(2,2.0)
    whitenoise = 0.
    order = 1

    # Define training x data
    training_x_loc = np.arange(train_width)
    training_y_loc = np.arange(train_height)
    training_x_loc, training_y_loc = np.meshgrid(training_x_loc, training_y_loc)
    training_x_loc = training_x_loc.flatten()
    training_y_loc = training_y_loc.flatten()
    X = np.column_stack([training_x_loc, training_y_loc])
    # Define Y_err
    Y_err = np.zeros(X.shape[0])

    # Define sample data
    eval_x_loc = np.arange(sample_width) * (train_width / sample_width)
    eval_y_loc = np.arange(sample_height) * (train_height / sample_height)
    eval_x_loc, eval_y_loc = np.meshgrid(eval_x_loc, eval_y_loc)
    eval_x_loc = eval_x_loc.flatten()
    eval_y_loc = eval_y_loc.flatten()
    X_eval = np.column_stack((eval_x_loc, eval_y_loc)).astype(float)
    
    # Define block width / height
    block_width = int(np.floor(width / train_width))
    block_height = int(np.floor(height / train_height))
    block_size = block_height * block_height

    # Define block "stamp"
    block_x = np.arange(block_width).astype(int)
    block_y = np.arange(block_height).astype(int)
    block_x, block_y = np.meshgrid(block_x, block_y)
    block_x = block_x.flatten()
    block_y = block_y.flatten()

    # Define Y data
    Y_red       = []
    Y_green     = []
    Y_blue      = []
    Y_gray      = []
    for (i, j) in X:
        x_start = i * block_width
        y_start = j * block_height
        x = block_x + x_start
        y = block_y + y_start
        reds, blues, greens = 0, 0, 0
        for (xi, yi) in zip(x,y):
            r, g, b = pixels[xi, yi]
            reds += r
            greens += g
            blues += b
        Y_red.append(   reds    / block_size)
        Y_blue.append(  blues   / block_size)
        Y_green.append( greens  / block_size)
        Y_gray.append((reds + blues + greens) / (3 * block_size))

    Y_red = np.asarray(Y_red)
    Y_green = np.asarray(Y_green)
    Y_blue = np.asarray(Y_blue)
    Y_gray = np.asarray(Y_gray)
    # Change X to float
    X = X.astype(float)

    toc = time.perf_counter()
    print(f"Setting up training data time: {toc-tic:.6f} s!")
    # Grayscale
    print(f"Training grayscale GP on {train_width} by {train_height} grid!")
    tic = time.perf_counter()
    gp = HybridGP(
        X,
        Y_gray,
        Y_err,
        scale,
        whitenoise,
        order,
    )
    toc = time.perf_counter()
    print(f"time: {toc-tic:.6f} s")
    # Eval gray
    print(f"Evaluating grayscale GP on {sample_width} by {sample_height} grid!")
    tic = time.perf_counter()
    y_eval = gp.predict_mean(X_eval)
    toc = time.perf_counter()
    print(f"time: {toc-tic:.6f} s")
    y_img = y_eval.reshape((sample_height, sample_width))
    plt.imshow(y_img, cmap='gray')
    plt.savefig("test_Nighthawks_gray.png")
    plt.close()

if __name__ == "__main__":
    test_Nighthawks()



#!/usr/bin/env python3
######## Imports ########
#### Standard Library ####
import argparse
import subprocess
import shutil
import os
import sys
from pathlib import Path

######## Arguments ########
# Define parser object
parser = argparse.ArgumentParser()
# Add arguments
parser.add_argument('--manifest-path', required=True)
parser.add_argument('--output', required=True)
parser.add_argument('--lib-name', required=True)
# Collect arguments
args = parser.parse_args()

######## Execution ########
env = os.environ.copy()
if sys.platform == "darwin":
    current_flags = env.get("RUSTFLAGS", "")
    macos_flags = "-C link-arg=-undefined -C link-arg=dynamic_lookup"
    env["RUSTFLAGS"] = f"{current_flags} {macos_flags}"

# Check that Cargo.toml file is valid
subprocess.check_call(
    ['cargo', 'metadata', '--no-deps', '--manifest-path', args.manifest_path],
    stdout=subprocess.DEVNULL,
)

# compile rust project
subprocess.check_call(
    [
        'cargo', 'build',
        '--release',
        '--manifest-path', args.manifest_path,
    ],
    env=env,
)

manifest_dir = Path(args.manifest_path).parent
target_dir = manifest_dir / 'target' / 'release'
crate_name = manifest_dir.name
cargo_prefix = f"lib{crate_name}"

# Platform dependent, but PyO3 naming is consistent
for file in target_dir.iterdir():
    if crate_name in file.name and file.suffix in ('.so', '.dylib', '.pyd'):
        shutil.copy(file, args.output)
        break
else:
    raise RuntimeError("Built library not found")





"""
* `new` - Constructor for Hybrid/Inner GP
* `predict_mean` - Predict GP interpolation mean on sample array
* `predict_mean_sgl` - Predict GP interpolation mean on single sample
* `predict_var` - Predict GP interpolation variance on sample array
* `predict_var_sgl` - Predict GP interpolation variance on single sample
* `predict_cov_sgl` - Predict GP covariance between two sample points
* `retrain` - Retrain GP with new hyperparameters
* `log_marginal_likelihood` - Evaluate the log marginal likelihood of the model
* `ndim` - Return the number of dimensions for the GP model
* `whitenoise` - Return the whitenoise hyperparameter for the GP model
* `to_hdf5` - Save the GP to HDF5
* `from_hdf5` - Load a GP from HDF5
* `generate_ldl` - Generate the LDL decomposition for a loaded model
"""
def test_sine():
    #### Imports #### 
    import time
    import numpy as np
    import matplotlib
    from matplotlib import pyplot as plt
    from csrk import HybridGP
    #### Setup ####
    xmin, xmax = 0, 16*np.pi
    ntrain = 100
    neval = 101
    X = np.atleast_2d(np.linspace(xmin,xmax,ntrain)).T
    Y = np.sin(X.flatten())
    Y_err = np.zeros_like(Y)
    scale = np.asarray([1.])
    whitenoise = 1e-9
    order = 3

    #### Train the GP ####
    gp = HybridGP(X, Y, Y_err, scale, whitenoise, order)
    #### Sample the GP ####
    x = np.atleast_2d(np.linspace(xmin,xmax,neval)).T
    # Test predict mean and get values
    y = gp.predict_mean(x)
    # Test predict_mean_sgl
    assert np.isclose(y[0], gp.predict_mean_sgl(x[0]))
    # Test predict_var
    var = gp.predict_var(x)
    # Test predict_var_sgl
    assert np.isclose(var[0], gp.predict_var_sgl(x[0]))
    # Test predict_cov_sgl
    cov_0_1 = gp.predict_cov_sgl(x[0], x[1])
    # Test retrain
    gp.retrain(2*scale,2*whitenoise,order)
    y2 = gp(x)
    # Estimate difference
    diff = np.abs(y - y2)
    assert diff.max() < 0.2

    #### Serialization ####
    gp.to_hdf5("test_serialization.hdf5", "sine_gp")
    gp2 = HybridGP.from_hdf5("test_serialization.hdf5", "sine_gp")
    assert gp2.whitenoise == gp.whitenoise
    assert gp2.ndim == 1
    assert np.allclose(gp2.scale, gp.scale)
    assert np.allclose(gp2.predict_mean(x), y2)
    #### Plot the result ####
    plt.plot(X.flatten(),Y, label="training")
    plt.plot(x,y, label="eval", linestyle='dashed')
    plt.legend()
    plt.savefig("test_csrk_sine.png")

if __name__ == "__main__":
    test_sine()



######## Setup ########
FNAME_SNR_TRAINING = "originals/SNR_training.csv"
NDIM = 2
ORDER = 1
WHITENOISE = 0.
SCALE_K = 2.

######## Tests ########
def test_SNR_training():
    import time
    import numpy as np
    from csrk import HybridGP
    
    ## load training data
    alldata = np.loadtxt(FNAME_SNR_TRAINING, delimiter=",")
    assert alldata.shape[1] == NDIM + 1
    print(alldata.strides)
    X = alldata[:,:-1]
    print(X.strides)
    print(X.dtype)
    Y = alldata[:,-1]
    Y_err = np.zeros_like(Y)
    msq_order = ORDER
    whitenoise = WHITENOISE

    ## Determine scale
    # Look at the first point
    x0 = alldata[0,:-1]
    # Find the distances between the first point and all other points
    dists = np.abs(X - x0)
    for dim in range(NDIM):
        mask = ~np.isclose(dists[:,dim], 0)
        dists = dists[mask]
    scale_1 = np.min(dists, axis=0)
    scale = scale_1 * SCALE_K

    # Grayscale
    print(f"Training SNR GP ({Y.size} pts)!")
    tic = time.perf_counter()
    gp = HybridGP(
        X,
        Y,
        Y_err,
        scale,
        whitenoise,
        msq_order,
    )
    toc = time.perf_counter()
    print(f"time: {toc-tic:.6f} s")
    return

if __name__ == "__main__":
    test_SNR_training()
